package com.simonstl.fragment;

import org.xml.sax.*;
import org.xml.sax.helpers.AttributesImpl;
import org.xml.sax.helpers.DefaultHandler;
import org.xml.sax.XMLFilter;
import org.xml.sax.helpers.XMLFilterImpl;
import org.apache.xerces.utils.regex.RegularExpression;
import org.apache.xerces.utils.regex.Match;
import java.util.*;

/*
The contents of this file are subject to the Mozilla Public License  Version 1.1 (the "License"); you may not use this file except in  compliance with the License. You may obtain a copy of the License at 
http://www.mozilla.org/MPL/ 

Software distributed under the License is distributed on an "AS IS"  basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the  License for the specific language governing rights and limitations under  the License. 

The Original Code is available at  http://simonstl.com/projects/fragment/original. 

The Initial Developer of the Original Code is Simon St.Laurent. Portions  created by Simon St.Laurent are Copyright (C) 2001. All Rights Reserved. 

Contributor(s): 
*/

/**
* This class provides a SAX 2.0 Filter which uses regular expressions to
* fragment particular elements into smaller labelled components.
* This class relies on the regular expression support built into the
* Apache Xerces parser, notably:
* org.apache.xerces.utils.regex.RegularExpression
* org.apache.xerces.utils.regex.Match
* 
* The RulesLoader class provides support for loading the rules files necessary
* to make the FragmentFilter work.
* @version 0.02  21 June 2001
* @author Simon St.Laurent
*/

public class FragmentFilter extends XMLFilterImpl { 

    private FragmentRules rules;   //set of rules
    private BoolStack inMatch;       //keeps track of whether the current element is a match
    private BufferStack buffer;   //buffers text before fragmentation
    private AttributesImpl emptyAtts; //empty attribute list
/**
* An empty constructor that requires the use of setParent before starting
* I don't think this works.
*/
public FragmentFilter () {
    emptyAtts=new AttributesImpl();
    inMatch=new BoolStack();
    buffer=new BufferStack();
 } 

/**
* A constructor that takes the parser which will feed it SAX events
*/

public FragmentFilter (XMLReader parent) {
    super(parent); 
    emptyAtts=new AttributesImpl();
    inMatch=new BoolStack();
    buffer=new BufferStack();

} 

/**
* sets up the rules, read in from a config file by RulesLoader or
* otherwise concocted in Java
* @param newRules  the set of rules the FragmentFilter will apply to content
*/
public void setRules (FragmentRules newRules) {
  rules=newRules;
}

/**
* in case you ever need to get the rules out of FragmentFilter
*/
public FragmentRules getRules () {
  return rules;
}

/** 
* Check at startElement for elements which should be fragmented. 
* if the element's namespace URI and local name match a rule, text
* will be suppressed until the end element event.
*/ 

public void startElement (String uri, String localName, String qName, Attributes atts) throws SAXException {
 //
 boolean ruled=rules.isRule(uri, localName);
 if (ruled) {
//System.out.println("match on: "+localName);
  inMatch.push(true);
  buffer.newEntry();
 } else {
  inMatch.push(false);
 }
//System.out.println(inMatch.peek());
 super.startElement(uri, localName, qName, atts); 
} 

/**
* If the current context isn't matched by any rules, this passes text through.
* Otherwise, it stores it in a buffer.  The strange approach to the buffering is
* dictated by Xerces' passing enormously excessive character arrays.
*/

public void characters (char[] ch, int start, int len)  throws SAXException {
String chString=new String (ch);
String localString=chString.substring (start, start+len);
 if (inMatch.peek()) {
  buffer.append(localString);
 }
 else { 
   super.characters (ch, start, len);
 }
}

/** 
* At the end of the element, break up its content using the regex and report all of it to the recipient.
* Note: right now, this method is just calling this class's own methods.  That puts a limit on recursive
* processing for fragmentation, and is a limitation I hope to remove in the near future.
* Also note that this approach only works for text stored atomically - don't try it on mixed content.
* Results will be ugly. 
*/

public void endElement (String uri, String localName, String qName) throws SAXException {

 if (inMatch.peek()) {

//System.out.println("buffer contains: "+ buffer.toString());
   String matchString=buffer.toString();
   FragmentRule currentRule=rules.getRule(uri,localName);
   String regEx=currentRule.matchRegEx;
   ArrayList results=currentRule.results;
//System.out.println("regex: "+regEx);
   RegularExpression regExProc=new RegularExpression(regEx);
   Match match=new Match();
   if (regExProc.matches(matchString, match)) {
     int numberMatches=match.getNumberOfGroups();
//System.out.println(numberMatches-1+" matches");
     for (int i=1; i<numberMatches; i++) {
       String value=match.getCapturedText(i);
       Element currentElement=(Element)results.get(i-1);
       String elemURI=currentElement.nsURI;
       String elemLocalName=currentElement.localName;
       String elemQName=elemLocalName;
       if (!(currentElement.prefix.equals(""))) {
         elemQName=currentElement.prefix+":"+elemLocalName;
       }
//System.out.println(i+". "+value);
      this.startElement(elemURI, elemLocalName, elemQName, emptyAtts);
      this.characters(value.toCharArray(), 0, value.length());
      this.endElement(elemURI, elemLocalName, elemQName);

     }
        inMatch.pop();
//      buffer.pop();
   } else {
System.out.println("False claim of match!!");
     buffer.pop();
     inMatch.pop();
      this.characters(matchString.toCharArray(), 0, matchString.length());
   } 


 }else {
        inMatch.pop();
   }
 super.endElement(uri, localName, qName);
 } 


} 
